import os

import np_config
import np_logging

logger = np_logging.get_logger(__name__)

CONFIG: dict = np_config.fetch('/projects/np_tools/config')
"""Package config dict with paths etc."""

ON_WINDOWS: bool = os.name == 'nt'
"""True if running on Windows."""



import pytest
import np_logging

from np_tools import *

TEST_BYTES = b'test'
EMPTY_BYTES = b''

@pytest.fixture
def src_dir(tmp_path):
    src = tmp_path / 'src'
    src.mkdir()
    return src

@pytest.fixture
def dest_dir(tmp_path):
    dest_dir = tmp_path / 'dest'
    dest_dir.mkdir()
    return dest_dir

@pytest.fixture
def src_file(src_dir):
    src_file = src_dir / 'test.txt'
    src_file.write_bytes(TEST_BYTES)
    return src_file

@pytest.fixture
def empty_dest_file(dest_dir, src_file):
    empty_dest_file = dest_dir / src_file.name
    empty_dest_file.write_bytes(EMPTY_BYTES)
    return empty_dest_file

def dest_matches_src(src, dest):
    if not src.is_dir():
        return dest.read_bytes() == src.read_bytes()
    for src_path in src.rglob('*'):
        dest_path = dest / src_path.relative_to(src)
        assert dest_path.exists()
        if src_path.is_dir():
            continue
        assert dest_path.read_bytes() == src_path.read_bytes()
    return True

def test_copy_dir_to_dir(src_file, dest_dir):
    copy(src_file.parent, dest_dir)
    assert dest_matches_src(src_file.parent, dest_dir)
    
def test_copy_file_to_dir(dest_dir, src_file):
    copy(src_file, dest_dir)
    assert dest_matches_src(src_file.parent, dest_dir)

def test_copy_file_to_file(src_file, dest_dir):
    copy(src_file, dest_dir / src_file.name)
    assert dest_matches_src(src_file.parent, dest_dir)

def test_copy_dir_to_non_existant_dir(src_file, tmp_path):
    dest = tmp_path / 'dest'
    copy(src_file, dest)
    assert dest_matches_src(src_file.parent, dest)
    
def test_copy_file_to_existing_file(src_file, empty_dest_file):
    copy(src_file, empty_dest_file.parent)
    assert dest_matches_src(src_file.parent, empty_dest_file.parent)
    
def test_move_dir(src_file, dest_dir):
    try:
        move(src_file.parent, dest_dir)
    except PermissionError:
        pass
    else:
        assert not src_file.parent.exists()
        assert (dest_dir / src_file.name).read_bytes() == TEST_BYTES
    
def test_move_file(src_file, dest_dir):
    try:
        move(src_file, dest_dir)
    except PermissionError:
        pass
    else:
        assert not src_file.exists()
        assert src_file.parent.exists()
        assert (dest_dir / src_file.name).read_bytes() == TEST_BYTES


from __future__ import annotations

import pathlib
import tempfile
from typing import Optional
import uuid

import np_logging
import pynwb

logger = np_logging.getLogger(__name__)


def init_nwb(
    session: np_session.Session,
    description: str = 'Data and metadata for a Neuropixels ecephys session',
) -> pynwb.NWBFile:
    """
    Init `NWBFile` with minimum required arguments from an
    `np_session.Session` instance.
    """
    return pynwb.NWBFile(
        session_description=description,
        identifier=str(uuid.uuid4()),  # globally unique for this nwb - not human-readable
        session_start_time=session.start,
    )
    

def load_nwb(
    nwb_path: str | pathlib.Path,
    ) -> pynwb.NWBFile:
    """Load `pynb.NWBFile` instance from path."""
    logger.info(f'Loading .nwb file at {nwb_path}')
    with pynwb.NWBHDF5IO(nwb_path, mode='r') as f:
        return f.read()


def save_nwb(
    nwb_file: pynwb.NWBFile,
    output_path: Optional[str | pathlib.Path] = None,
    ) -> pathlib.Path:
    """
    Write `pynb.NWBFile` instance to disk.
    
    Temp dir is used if `output_path` isn't provided.
    """
    if output_path is None:
        output_path = pathlib.Path(tempfile.mkdtemp()) / f'{nwb_file.session_id}.nwb'
    
    nwb_file.set_modified()
    # not clear if this is necessary, but suggested by docs:
    # https://pynwb.readthedocs.io/en/stable/_modules/pynwb.html

    logger.info(f'Writing .nwb file `{nwb_file.session_id!r}` to {output_path}')
    with pynwb.NWBHDF5IO(output_path, mode='w') as f:
        f.write(nwb_file, cache_spec=True)
    logger.debug(f'Writing complete for nwb file `{nwb_file.session_id!r}`')
    return output_path



"""Tools for working with Open Ephys raw data files."""
from __future__ import annotations

import doctest
import json
import pathlib
import tempfile
from typing import Any, Generator, Optional, Sequence, Iterable

import np_config
import np_logging

import np_tools.file_io as file_io


logger = np_logging.get_logger(__name__)

DEFAULT_PROBES = 'ABCDEF'


def is_new_ephys_folder(path: pathlib.Path) -> bool:
    """Look for any hallmarks of a v0.6.x Open Ephys recording in path or subfolders."""

    globs = (
        'Record Node*',
        'structure*.oebin',
    )
    components = tuple(_.replace('*', '') for _ in globs)

    if any(_.lower() in path.as_posix().lower() for _ in components):
        return True

    for glob in globs:
        if next(path.rglob(glob), None):
            return True
    return False

def correct_structure_oebin_files(dest: pathlib.Path) -> None:
    """
    Overwrites oebin files with correct metadata, removing entries for missing
    folders.
    
    - in cases of probes not being inserted, raw data may be removed or not copied
      over, and the structure.oebin will contain data paths for entries that don't exist
    - this function will remove those entries from the structure.oebin file
    """
    logger.debug('Checking structure.oebin for missing folders...')
    recording_dirs = dest.rglob('recording[0-9]*')
    for recording_dir in recording_dirs:
        if not recording_dir.is_dir():
            continue
        oebin_path = recording_dir / 'structure.oebin'
        if not (oebin_path.is_symlink() or oebin_path.exists()):
            logger.warning(f'No structure.oebin found in {recording_dir}')
            continue
        logger.debug(f'Examining oebin: {oebin_path} for correction')
        if oebin_path.is_symlink():
            oebin_path = np_config.normalize_path(oebin_path.readlink())
        oebin_obj = read_oebin(oebin_path)
        any_removed = False
        for subdir_name in ('events', 'continuous'):    
            subdir = oebin_path.parent / subdir_name
            # iterate over copy of list so as to not disrupt iteration when elements are removed
            for device in [device for device in oebin_obj[subdir_name]]:
                if not (subdir / device['folder_name']).exists():
                    logger.info(f'{device["folder_name"]} not found in {subdir}, removing from structure.oebin')
                    oebin_obj[subdir_name].remove(device)
                    any_removed = True
        if any_removed:
            oebin_path.unlink()
            oebin_path.write_text(json.dumps(oebin_obj, indent=4))
            logger.debug('Overwrote structure.oebin with corrected strcuture.oebin')
            
def is_complete_ephys_folder(path: pathlib.Path) -> bool:
    """Look for all hallmarks of a complete v0.6.x Open Ephys recording."""
    # TODO use structure.oebin to check for completeness
    if not is_new_ephys_folder(path):
        return False
    for glob in ('continuous.dat', 'spike_times.npy', 'spike_clusters.npy'):
        if not next(path.rglob(glob), None):
            logger.debug(f'Could not find {glob} in {path}')
            return False
    return True


def is_valid_ephys_folder(
    path: pathlib.Path, min_size_gb: Optional[int | float] = None,
) -> bool:
    """Check a single dir of raw data for size and v0.6.x+ Open Ephys."""
    if not path.is_dir():
        return False
    if not is_new_ephys_folder(path):
        return False
    if min_size_gb is not None and file_io.dir_size(path) < min_size_gb * 1024**3: # GB
        return False
    return True


def get_ephys_root(path: pathlib.Path) -> pathlib.Path:
    """Returns the parent of the first `Record Node *` found in the supplied
    path.

    >>> get_ephys_root(pathlib.Path('A:/test/Record Node 0/Record Node test')).as_posix()
    'A:/test'
    """
    if 'Record Node' not in path.as_posix():
        raise ValueError(
            f"Could not find 'Record Node' in {path} - is this a valid raw ephys path?"
        )
    return next(
        p.parent
        for p in path.parents
        if 'Record Node'.lower() in p.name.lower()
    )



def get_filtered_ephys_paths_relative_to_record_node_parents(
    toplevel_ephys_path: pathlib.Path,
    specific_recording_dir_names: Iterable[str] | None = None
    ) -> Generator[tuple[pathlib.Path, pathlib.Path], None, None]:
    """For restructuring the raw ephys data in a session folder, we want to
    discard superfluous recording folders and only keep the "good" data, but
    with the overall directory structure relative to `Record Node*` folders intact.
    
    Supply a top-level path that contains `Record Node *`
    subfolders somewhere in its tree. 
    
    Returns a generator akin to `path.rglob('Record Node*')` except:
    - only paths associated with the "good" ephys data are returned (with some
    assumptions made about the ephys session)
    - a tuple of two paths is supplied:
        - `(abs_path, abs_path.relative_to(record_node.parent))`
        ie. path[1] should always start with `Record Node *`
    
    Expectation is:
    - `/npexp_path/ephys_*/Record Node */ recording1 / ... / continuous.dat`
    
    ie. 
    - one recording per `Record Node *` folder
    - one subfolder in `npexp_path/` per `Record Node *` folder (following the
    pipeline `*_probeABC` / `*_probeDEF` convention for recordings split across two
    drives)
    

    Many folders have:
    - `/npexp_path/Record Node */ recording*/ ...` 
    
    ie.
    - multiple recording folders per `Record Node *` folder
    - typically there's one proper `recording` folder: the rest are short,
    aborted recordings during setup
    
    We filter out the superfluous small recording folders here.
    
    Some folders (Templeton) have:
    - `/npexp_path/Record Node */ ...`
    
    ie.
    - contents of expected ephys subfolders directly deposited in npexp_path
    
    >>> path = pathlib.Path("//allen/programs/mindscope/workgroups/dynamicrouting/PilotEphys/Task 2 pilot/DRpilot_686176_20231206/DRpilot_686176_20231206")
    >>> len(tuple(get_filtered_ephys_paths_relative_to_record_node_parents(path)))
    230
    >>> len(tuple(get_filtered_ephys_paths_relative_to_record_node_parents(path, 'recording1')))
    230
    >>> len(tuple(get_filtered_ephys_paths_relative_to_record_node_parents(path, 'recording2')))
    230
    >>> len(tuple(get_filtered_ephys_paths_relative_to_record_node_parents(path, ('recording1', 'recording2'))))
    456
    """
    def is_recording_name(name: str) -> bool:
        return name.startswith('recording') and name[9:].isdigit()
    
    if specific_recording_dir_names:
        if isinstance(specific_recording_dir_names, str):
            specific_recording_dir_names = (specific_recording_dir_names,)
        for name in specific_recording_dir_names:
            if not is_recording_name(name):
                raise ValueError(f'specific_recording_dir_names must be in the format "recording1": got {name!r}')

    record_nodes = toplevel_ephys_path.rglob('Record Node*')
    
    if specific_recording_dir_names:
        for record_node in record_nodes:
            for abs_path in record_node.rglob('*'):          
                is_recording_child = any(is_recording_name(name) for name in abs_path.parts)
                is_matching_name = any(name in abs_path.parts for name in specific_recording_dir_names)
                if is_recording_child and not is_matching_name:
                    continue
                yield abs_path, abs_path.relative_to(record_node.parent) 
    else: 
        for record_node in record_nodes:
            superfluous_recording_dirs = tuple(
                _.parent for _ in get_superfluous_oebin_paths(record_node)
            )
            logger.debug(f'Found {len(superfluous_recording_dirs)} superfluous recording dirs to exclude: {superfluous_recording_dirs}')
            
            for abs_path in record_node.rglob('*'):
                is_superfluous_path = abs_path in superfluous_recording_dirs or any(_ in abs_path.parents for _ in superfluous_recording_dirs)
                
                if is_superfluous_path:
                    continue
                
                yield abs_path, abs_path.relative_to(record_node.parent)
        
       
def get_raw_ephys_subfolders(
    path: pathlib.Path, min_size_gb: Optional[int | float] = None
) -> tuple[pathlib.Path, ...]:
    """
    Return raw ephys recording folders, defined as the root that Open Ephys
    records to, e.g. `A:/1233245678_366122_20220618_probeABC`.
    """ 

    subfolders = set()

    for f in pathlib.Path(path).rglob('continuous.dat'):

        if any(
            k in f.as_posix().lower()
            for k in [
                'sorted',
                'extracted',
                'curated',
            ]
        ):
            # skip sorted/extracted folders
            continue

        subfolders.add(get_ephys_root(f))

    if min_size_gb is not None:
        subfolders = {_ for _ in subfolders if file_io.dir_size(_) < min_size_gb * 1024**3}

    return tuple(sorted(list(subfolders), key=lambda s: str(s)))


# - If we have probeABC and probeDEF raw data folders, each one has an oebin file:
#     we'll need to merge the oebin files and the data folders to create a single session
#     that can be processed in parallel
def get_single_oebin_path(path: pathlib.Path) -> pathlib.Path:
    """Get the path to a single structure.oebin file in a folder of raw ephys data.

    - There's one structure.oebin per `recording*` folder
    - Raw data folders may contain multiple `recording*` folders
    - Datajoint expects only one structure.oebin file per Session for sorting
    - If we have multiple `recording*` folders, we assume that there's one
        good folder - the largest - plus some small dummy / accidental recordings
    """
    if not path.is_dir():
        raise ValueError(f'{path} is not a directory')

    oebin_paths = tuple(path.rglob('structure*.oebin'))
        
    if not oebin_paths:
        raise FileNotFoundError(f'No structure.oebin file found in {path}')

    if len(oebin_paths) == 1:
        return oebin_paths[0]
    
    oebin_parents = (_.parent for _ in oebin_paths)
    dir_sizes = tuple(file_io.dir_size(_) for _ in oebin_parents)
    return oebin_paths[dir_sizes.index(max(dir_sizes))]


def get_superfluous_oebin_paths(path: pathlib.Path) -> tuple[pathlib.Path, ...]:
    """Get the paths to any oebin files in `recording*` folders that are not
    the largest in a folder of raw ephys data. 
    
    Companion to `get_single_oebin_path`.
    """
    
    all_oebin_paths = tuple(path.rglob('structure*.oebin'))
    
    if len(all_oebin_paths) == 1:
        return tuple()
    
    return tuple(set(all_oebin_paths) - {(get_single_oebin_path(path))})


def assert_xml_files_match(paths: Sequence[pathlib.Path]) -> None:
    """Check that all xml files are identical, as they should be for
    recordings split across multiple locations e.g. A:/*_probeABC, B:/*_probeDEF
    or raise an error.
    
    Update: xml files on two nodes can be created at slightly different times, so their `date`
    fields may differ. Everything else should be identical.
    """
    if not all(s == '.xml' for s in [p.suffix for p in paths]):
        raise ValueError('Not all paths are XML files')
    if not all(p.is_file() for p in paths):
        raise FileNotFoundError(
            'Not all paths are files, or they do not exist'
        )
    if not file_io.checksums_match(paths):
        
        # if the files are the same size and were created within +/- 1 second
        # of each other, we'll assume they're the same
        
        created_times = tuple(p.stat().st_ctime for p in paths)
        created_times_equal = all(created_times[0] - 1 <= t <= created_times[0] + 1 for t in created_times[1:])
        
        sizes = tuple(p.stat().st_size for p in paths)
        sizes_equal = all(s == sizes[0] for s in sizes[1:])
        
        if not (sizes_equal and created_times_equal):
            raise AssertionError('XML files do not match')


def get_merged_oebin_file(
    paths: Sequence[pathlib.Path], exclude_probe_names: Optional[Sequence[str]] = None
) -> pathlib.Path:
    """Merge two or more structure.oebin files into one.

    For recordings split across multiple locations e.g. A:/*_probeABC,
    B:/*_probeDEF
    - if items in the oebin files have 'folder_name' values that match any
    entries in `exclude_probe_names`, they will be removed from the merged oebin
    """
    if isinstance(paths, pathlib.Path):
        return paths
    if any(not p.suffix == '.oebin' for p in paths):
        raise ValueError('Not all paths are .oebin files')
    if len(paths) == 1:
        return paths[0]

    # ensure oebin files can be merged - if from the same exp they will have the same settings.xml file
    assert_xml_files_match(
        [p / 'settings.xml' for p in [o.parent.parent.parent for o in paths]]
    )

    logger.debug(f'Creating merged oebin file from {paths}')
    merged_oebin = dict()
    for oebin_path in sorted(paths):
        oebin_data = read_oebin(oebin_path)

        for key in oebin_data:

            # skip if already in merged oebin
            if merged_oebin.get(key) == oebin_data[key]:
                continue

            # 'continuous', 'events', 'spikes' are lists, which we want to concatenate across files
            if isinstance(oebin_data[key], list):
                for item in oebin_data[key]:
                    
                    # skip if already in merged oebin
                    if item in merged_oebin.get(key, []):
                        continue
                    
                    # skip probes in excl list (ie. not inserted)
                    if exclude_probe_names and any(
                        p.lower() in item.get('folder_name', '').lower()
                        for p in exclude_probe_names
                    ):
                        continue
                    
                    # insert in merged oebin
                    merged_oebin.setdefault(key, []).append(item)

    if not merged_oebin:
        raise ValueError('No data found in structure.oebin files')
    
    merged_oebin_path = pathlib.Path(tempfile.gettempdir()) / 'structure.oebin'
    merged_oebin_path.write_text(json.dumps(merged_oebin, indent=4))
    return merged_oebin_path


def read_oebin(path: str | pathlib.Path) -> dict[str, Any]:
    return json.loads(pathlib.Path(path).read_bytes())


if __name__ == '__main__':
    doctest.testmod()



"""
Tools for running commands via ssh on remote hosts.
"""
from __future__ import annotations


import doctest
import subprocess
import sys
from typing import Any

import fabric
import np_config
import np_logging
from np_tools.config import ON_WINDOWS
from typing_extensions import Literal

logger = np_logging.get_logger(__name__)

HPC_CREDENTIALS: dict[Literal['user', 'password'], str] = np_config.fetch(
    '/logins'
)['svc_neuropix']

NP_COMPUTERS: dict[str, str]


def ssh(host: str) -> fabric.Connection:
    """Fabric connection to `host` using `svc_neuropix` creds.

    >>> with ssh('w10svad0139') as connection:
    ...     response = connection.run('echo "hello world"', hide=True)

    """
    return fabric.Connection(
        host=host,
        user=HPC_CREDENTIALS['user'],
        connect_kwargs=dict(password=HPC_CREDENTIALS['password']),
    )


hpc = ssh('hpc-login')
"""Fabric connection to `hpc-login` using `svc_neuropix` creds."""


def run_cmd_on_host(
    host: str,
    cmd: str,
    *,
    hide_output: bool = True,  # suppresses stdout
    warn: bool = True,  # suppresses bad exit code error
) -> Any:
    """Run `cmd` on `host` command-line via ssh.

    If `host` is 'localhost', run `cmd` locally in a subprocess.

    >>> print(run_cmd_on_host('w10svad0139', 'hostname').stdout.strip())
    W10SVAD0139
    >>> run_cmd_on_host('localhost', 'echo hello world').stdout.strip()
    b'hello world'
    >>> _ = run_cmd_on_host('localhost', 'echo hello world', hide_output=False)

    # prints 'hello world'
    """
    if host == 'localhost':
        logger.debug('Running command on localhost: %r', cmd)
        result = subprocess.run(
            cmd,
            shell=True,
            check=warn,
            capture_output=hide_output,
        )
        logger.debug('Return code: %s', result.returncode)
        return result

    with ssh(host) as connection:
        logger.debug(
            'Sending command to %s via fabric as %s: %r',
            connection.host,
            connection.user,
            cmd,
        )
        result = connection.run(cmd, hide=hide_output, warn=warn)
        logger.debug('Return code: %s', result.return_code)
    return result


def assert_cli_tool_on_host(host: str, executable: str) -> None:
    """Assert that `executable` exists and can be accessed on `host` via ssh.

    >>> if ON_WINDOWS: assert_cli_tool_on_host('localhost', "robocopy") == None
    True
    >>> if ON_WINDOWS: assert_cli_tool_on_host('w10svad0139', "robocopy") == None
    True

    >>> assert_cli_tool_on_host('localhost', "not-robocopy")
    Traceback (most recent call last):
    ...
    AssertionError
    >>> assert_cli_tool_on_host('w10svad0139', "not-robocopy")
    Traceback (most recent call last):
    ...
    AssertionError

    """
    if host == 'localhost':
        return assert_cli_tool_on_localhost(executable)

    cmd = f'{executable} /?'

    response = run_cmd_on_host(host, cmd, hide_output=True, warn=True)
    if response.return_code == 1:
        raise AssertionError(f'{cmd!r} failed on {host!r}.')


def assert_cli_tool_on_localhost(executable: str) -> None:
    """Assert that `executable` exists and can be accessed on local machine.

    >>> if ON_WINDOWS: assert_cli_tool_on_localhost("robocopy") == None
    True

    >>> assert_cli_tool_on_localhost("not-robocopy")
    Traceback (most recent call last):
    ...
    AssertionError

    """
    try:
        response = subprocess.run(
            [executable, '/?'],
            check=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except FileNotFoundError:
        if not ON_WINDOWS and executable == 'robocopy':
            raise AssertionError(
                f'robocopy is only available on Windows: running on {sys.platform}'
            )
        raise AssertionError(f'{executable!r} could not be found in PATH.')
    else:
        if response.returncode != 16:
            raise AssertionError(
                f'{executable!r} returned exit status {response.returncode}.'
            )


if __name__ == '__main__':
    doctest.testmod(verbose=True, optionflags=doctest.IGNORE_EXCEPTION_DETAIL)



"""
Tools for working with Open Ephys raw data files.
"""
from __future__ import annotations
import contextlib
import datetime
import os
import pathlib
import shutil
import subprocess
import time
from typing import Optional, Union

import crc32c
import np_config
import np_logging
import rich.progress

logger = np_logging.get_logger(__name__)

PathLike = Union[str, bytes, os.PathLike, pathlib.Path]
def from_pathlike(pathlike):
    return pathlib.Path(os.fsdecode(pathlike))

if os.name == 'nt':
    # Remote to remote symlink creation is disabled by default
    try:
        proc = subprocess.run('fsutil behavior set SymlinkEvaluation R2R:1', check=True)
    except subprocess.CalledProcessError:
        logger.info('Failed to enable remote-to-remote symlink creation: try running as admin')
        R2R_SYMLINKS_ENABLED = False
    else:
        R2R_SYMLINKS_ENABLED = True
else:
    R2R_SYMLINKS_ENABLED = True
    
def checksum(path: PathLike, show_progress_bar=True) -> str:
    path = from_pathlike(path)
    hasher = crc32c.crc32c
    def formatted(x):
        return f'{x:08X}'
    blocks_per_chunk = 4096
    multi_part_threshold_gb = 0.2
    if path.stat().st_size < multi_part_threshold_gb * 1024**3:
        return formatted(hasher(path.read_bytes()))
    hash = 0
    
    with open(path, 'rb') as f, get_progress():
        progress: rich.progress.Progress = globals()['progress']
        task = progress.add_task(f'Checksumming {path.name}', total=path.stat().st_size, visible=show_progress_bar)
        for chunk in iter(
            lambda: f.read(blocks_per_chunk), b''
        ):  
            progress.update(task, advance=blocks_per_chunk) # type: ignore
            hash = hasher(chunk, hash)
    progress.update(task, visible=False)
    return formatted(hash)

def get_progress() -> rich.progress.Progress | contextlib.nullcontext[None]:
    progress_context = contextlib.nullcontext()
    if 'progress' not in globals():
        progress_context = globals()['progress'] = rich.progress.Progress(
            rich.progress.TextColumn("{task.description}", justify="right"),
            rich.progress.BarColumn(),
            rich.progress.TimeRemainingColumn(),
            rich.progress.FileSizeColumn(),
            rich.progress.TotalFileSizeColumn(),
        )
    return progress_context

def get_copy_task(src) -> rich.progress.TaskID:
    get_progress()
    progress: rich.progress.Progress = globals()['progress']
    if not progress.tasks:
        task = progress.add_task(
            description='[cyan]Getting file sizes',
            start=False,
            )
        progress.update(
            task,
            description='[cyan]Copying files',
            total=size(src)
            )
        progress.start_task(task)
    return progress.tasks[0].id

def checksums_match(*paths: PathLike) -> bool:
    checksums = tuple(checksum(p) for p in paths)
    return all(c == checksums[0] for c in checksums)


def copy(src: PathLike, dest: PathLike, max_attempts: int = 2) -> None:
    """Copy `src` to `dest` with checksum validation.

    - copies recursively if `src` is a directory
    - if dest already exists, checksums are compared, copying is skipped if they match
    - attempts to copy up to 3 times if checksums don't match
    - replaces existing symlinks with actual files
    - creates parent dirs if needed
    """
    src, dest = from_pathlike(src), from_pathlike(dest)

    with get_progress():
        progress: rich.progress.Progress = globals()['progress']
        task = get_copy_task(src)
    
        if dest.exists() and dest.is_symlink():
            dest.unlink() # we'll replace symlink with src file
        
        if src.is_dir(): # copy files recursively
            for path in src.iterdir():
                copy(path, dest / path.name)
            return
        
        if not dest.suffix: # dest is a folder, but might not exist yet so can't use `is_dir`
            dest = dest / src.name
        dest.parent.mkdir(parents=True, exist_ok=True)
        
        if not dest.exists():
            shutil.copy2(src, dest)
            logger.debug(f'Copied {src} to {dest}')
            
        for _ in range(max_attempts):
            if checksums_match(src, dest):
                break
            shutil.copy2(src, dest)
        else:
            raise OSError(
                f'Failed to copy {src} to {dest} with checksum-validation after {max_attempts} attempts'
            )
        progress.update(task, advance=size(src))
        logger.debug(f'Copy of {src} at {dest} validated with checksum')


def move(src: PathLike, dest: PathLike, **rmtree_kwargs) -> None:
    """Copy `src` to `dest` with checksum validation, then delete `src`.
    """
    src, dest = from_pathlike(src), from_pathlike(dest)
    copy(src, dest)
    if src.is_dir():
        shutil.rmtree(src, **rmtree_kwargs)
    else:
        src.unlink()
    logger.debug(f'Deleted {src}')


def symlink(src: PathLike, dest: PathLike) -> None:
    """Create symlink at `dest` pointing to file at `src`.

    - creates symlinks recursively if `src` is a directory
    - creates parent dirs if needed (as folders, not symlinks)
    - skips if symlink already exists and points to `src`
    - replaces existing file or symlink pointing to a different location
    """
    src, dest = from_pathlike(src), from_pathlike(dest)
    if src.is_dir():
        for path in src.iterdir():
            symlink(src, dest / path.name)
    dest.parent.mkdir(parents=True, exist_ok=True)
    if dest.is_symlink() and dest.resolve() == src:
        logger.debug(f'Symlink already exists to {src} from {dest}')
        return
    with contextlib.suppress(FileNotFoundError):
        dest.unlink()
    with contextlib.suppress(FileExistsError):
        dest.symlink_to(src)
    logger.debug(f'Created symlink to {src} from {dest}')

    
def size(path: PathLike) -> int:
    """Return the size of a file or directory in bytes"""
    path = from_pathlike(path)
    return dir_size(path) if path.is_dir() else path.stat().st_size

def size_gb(path: PathLike) -> float:
    """Return the size of a file or directory in GB"""
    return round(size(path) / 1024 ** 3, 1)

def dir_size(path: PathLike) -> int:
    """Return the size of a directory in bytes"""
    path = from_pathlike(path)
    if not path.is_dir():
        raise ValueError(f'{path} is not a directory')
    dir_size = 0
    dir_size += sum(
        f.stat().st_size
        for f in path.rglob('*')
        if f.is_file()
    )
    return dir_size

def dir_size_gb(path: PathLike) -> float:
    """Return the size of a directory in GB"""
    return round(dir_size(path) / 1024 ** 3, 1)

def free_gb(path: PathLike) -> float:
    "Return free space at `path`, to .1 GB. Raises FileNotFoundError if `path` not accessible."
    path = from_pathlike(path)
    path = np_config.unc_to_local(path)
    return round(shutil.disk_usage(path).free / 1024 ** 3, 1)

def get_files_created_between(
    path: PathLike,
    glob: str = "*",
    start: float | datetime.datetime = 0,
    end: Optional[float | datetime.datetime] = None,
    reverse: bool = False,
) -> tuple[pathlib.Path, ...]:
    """Recusively get search for files in subfolders of `path` created between `start` and `end`.
    
    Sequence is sorted by ascending creation time (oldest first). `reverse` reverses this order.
    """
    path = from_pathlike(path)
    if not path.is_dir():
        raise ValueError(f'{path} is not a directory, cannot glob for files')
    if not end:
        end = time.time()
    start = start.timestamp() if isinstance(start, datetime.datetime) else start
    end = end.timestamp() if isinstance(end, datetime.datetime) else end
    def ctime(x):
        return x.stat().st_ctime
    files = (file for file in path.rglob(glob) if int(start) <= ctime(file) <= end)
    return tuple(sorted(files, key=ctime, reverse=reverse))
